from pathlib import Path

from bioimageio.core.resource_io import nodes, utils
from bioimageio.spec.shared import raw_nodes


def test_resolve_import_path(tmpdir):
    tmpdir = Path(tmpdir)
    manifest_path = tmpdir / "manifest.yaml"
    manifest_path.touch()
    source_file = Path("my_mod.py")
    (tmpdir / str(source_file)).write_text("class Foo: pass", encoding="utf8")
    node = raw_nodes.ImportableSourceFile(source_file=source_file, callable_name="Foo")
    uri_transformed = utils.UriNodeTransformer(root_path=tmpdir).transform(node)
    source_transformed = utils.SourceNodeTransformer().transform(uri_transformed)
    assert isinstance(source_transformed, nodes.ImportedSource)
    Foo = source_transformed.factory
    assert Foo.__name__ == "Foo"
    assert isinstance(Foo, type)


def test_resolve_directory_uri(tmpdir):
    node = raw_nodes.URI(Path(tmpdir).as_uri())
    uri_transformed = utils.UriNodeTransformer(root_path=Path(tmpdir)).transform(node)
    assert uri_transformed == Path(tmpdir)


def test_uri_available():
    pass  # todo


def test_all_uris_available():
    from bioimageio.core.resource_io.utils import all_sources_available

    not_available = {
        "uri": raw_nodes.URI(scheme="file", path="non_existing_file_in/non_existing_dir/ftw"),
        "uri_exists": raw_nodes.URI(scheme="file", path="."),
    }
    assert not all_sources_available(not_available)


def test_uri_node_transformer_is_ok_with_abs_path():
    from bioimageio.core.resource_io.utils import UriNodeTransformer

    # note: the call of .absolute() is required to add the drive letter for windows paths, which are relative otherwise
    tree = {"rel_path": Path("something/relative"), "abs_path": Path("/something/absolute").absolute()}
    assert not tree["rel_path"].is_absolute()
    assert tree["abs_path"].is_absolute()

    root = Path("/root").absolute()
    print(root)

    tree = UriNodeTransformer(root_path=root).transform(tree)
    assert tree["rel_path"].is_absolute()
    assert tree["rel_path"] == Path("/root/something/relative").absolute()
    assert tree["abs_path"].is_absolute()
    assert tree["abs_path"] == Path("/something/absolute").absolute()



import os
import bioimageio.spec as spec
from bioimageio.core import load_raw_resource_description, load_resource_description
from bioimageio.core.resource_io import nodes
from bioimageio.core.resource_io.utils import resolve_source
from bioimageio.core.resource_tests import test_model as _test_model
from marshmallow import missing


def _test_build_spec(
    spec_path,
    out_path,
    weight_type,
    tensorflow_version=None,
    opset_version=None,
    use_implicit_output_shape=False,
    add_deepimagej_config=False,
    use_original_covers=False,
    use_absoloute_arch_path=False,
    training_data=None,
    parent=None,
):
    from bioimageio.core.build_spec import build_model

    model_spec = load_raw_resource_description(spec_path, update_to_format="latest")
    root = model_spec.root_path
    assert isinstance(model_spec, spec.model.raw_nodes.Model)
    weight_source = model_spec.weights[weight_type].source

    cite = []
    for entry in model_spec.cite:
        entry_ = {"text": entry.text}
        has_url = entry.url is not missing
        has_doi = entry.doi is not missing
        assert has_url != has_doi
        if has_doi:
            entry_["doi"] = entry.doi
        else:
            entry_["url"] = entry.url
        cite.append(entry_)

    weight_spec = model_spec.weights[weight_type]
    dep_file = None if weight_spec.dependencies is missing else resolve_source(weight_spec.dependencies.file, root)
    if weight_type == "pytorch_state_dict":
        model_kwargs = None if weight_spec.kwargs is missing else weight_spec.kwargs
        architecture = str(weight_spec.architecture)
        if use_absoloute_arch_path:
            arch_path, cls_name = architecture.split(":")
            arch_path = os.path.abspath(os.path.join(root, arch_path))
            assert os.path.exists(arch_path)
            architecture = f"{arch_path}:{cls_name}"
        weight_type_ = None  # the weight type can be auto-detected
    elif weight_type == "torchscript":
        architecture = None
        model_kwargs = None
        weight_type_ = "torchscript"  # the weight type CANNOT be auto-detected
    else:
        architecture = None
        model_kwargs = None
        weight_type_ = None  # the weight type can be auto-detected

    authors = [{"name": auth.name, "affiliation": auth.affiliation} for auth in model_spec.authors]

    input_axes = [input_.axes for input_ in model_spec.inputs]
    output_axes = [output.axes for output in model_spec.outputs]
    preprocessing = [
        None
        if input_.preprocessing is missing
        else [{"name": preproc.name, "kwargs": preproc.kwargs} for preproc in input_.preprocessing]
        for input_ in model_spec.inputs
    ]
    postprocessing = [
        None
        if output.postprocessing is missing
        else [{"name": preproc.name, "kwargs": preproc.kwargs} for preproc in output.preprocessing]
        for output in model_spec.outputs
    ]

    kwargs = dict(
        weight_uri=weight_source,
        test_inputs=resolve_source(model_spec.test_inputs, root),
        test_outputs=resolve_source(model_spec.test_outputs, root),
        name=model_spec.name,
        description=model_spec.description,
        authors=authors,
        tags=model_spec.tags,
        license=model_spec.license,
        documentation=model_spec.documentation,
        dependencies=dep_file,
        cite=cite,
        root=model_spec.root_path,
        weight_type=weight_type_,
        input_axes=input_axes,
        output_axes=output_axes,
        preprocessing=preprocessing,
        postprocessing=postprocessing,
        output_path=out_path,
        add_deepimagej_config=add_deepimagej_config,
        maintainers=[{"github_user": "jane_doe"}],
        input_names=[inp.name for inp in model_spec.inputs],
        output_names=[out.name for out in model_spec.outputs],
    )
    if architecture is not None:
        kwargs["architecture"] = architecture
    if model_kwargs is not None:
        kwargs["model_kwargs"] = model_kwargs
    if tensorflow_version is not None:
        kwargs["tensorflow_version"] = tensorflow_version
    if opset_version is not None:
        kwargs["opset_version"] = opset_version
    if use_implicit_output_shape:
        kwargs["input_names"] = ["input"]
        kwargs["output_reference"] = ["input"]
        kwargs["output_scale"] = [[1.0, 1.0, 1.0, 1.0]]
        kwargs["output_offset"] = [[0.0, 0.0, 0.0, 0.0]]
    if add_deepimagej_config:
        kwargs["pixel_sizes"] = [{"x": 5.0, "y": 5.0}]
    if use_original_covers:
        kwargs["covers"] = resolve_source(model_spec.covers, root)
    if training_data is not None:
        kwargs["training_data"] = training_data
    if parent is not None:
        kwargs["parent"] = parent

    build_model(**kwargs)
    assert out_path.exists()
    loaded_model = load_resource_description(out_path)
    assert isinstance(loaded_model, nodes.Model)
    if add_deepimagej_config:
        loaded_config = loaded_model.config
        assert "deepimagej" in loaded_config

    if loaded_model.sample_inputs is not missing:
        for sample in loaded_model.sample_inputs:
            assert sample.exists()
    if loaded_model.sample_outputs is not missing:
        for sample in loaded_model.sample_outputs:
            assert sample.exists()

    assert loaded_model.maintainers[0].github_user == "jane_doe"

    attachments = loaded_model.attachments
    if attachments is not missing and attachments.files is not missing:
        for attached_file in attachments.files:
            assert attached_file.exists()

    # make sure there is one attachment per pre/post-processing
    if add_deepimagej_config:
        preproc, postproc = preprocessing[0], postprocessing[0]
        n_processing = 0
        if preproc is not None:
            n_processing += len(preproc)
        if postproc is not None:
            n_processing += len(postproc)
        if n_processing > 0:
            assert attachments.files is not missing
            assert n_processing == len(attachments.files)

    # test inference for the model to ensure that the weights were written correctly
    test_res = _test_model(out_path)
    assert test_res["error"] is None


def test_build_spec_pytorch(any_torch_model, tmp_path):
    _test_build_spec(any_torch_model, tmp_path / "model.zip", "pytorch_state_dict")


def test_build_spec_implicit_output_shape(unet2d_nuclei_broad_model, tmp_path):
    _test_build_spec(
        unet2d_nuclei_broad_model, tmp_path / "model.zip", "pytorch_state_dict", use_implicit_output_shape=True
    )


def test_build_spec_torchscript(any_torchscript_model, tmp_path):
    _test_build_spec(any_torchscript_model, tmp_path / "model.zip", "torchscript")


def test_build_spec_onnx(any_onnx_model, tmp_path):
    _test_build_spec(any_onnx_model, tmp_path / "model.zip", "onnx", opset_version=12)


def test_build_spec_keras(any_keras_model, tmp_path):
    _test_build_spec(
        any_keras_model, tmp_path / "model.zip", "keras_hdf5", tensorflow_version="1.12"
    )  # todo: keras for tf 2??


def test_build_spec_tf(any_tensorflow_model, tmp_path):
    _test_build_spec(
        any_tensorflow_model, tmp_path / "model.zip", "tensorflow_saved_model_bundle", tensorflow_version="1.12"
    )  # check tf version


def test_build_spec_tfjs(any_tensorflow_js_model, tmp_path):
    _test_build_spec(any_tensorflow_js_model, tmp_path / "model.zip", "tensorflow_js", tensorflow_version="1.12")


def test_build_spec_deepimagej(unet2d_nuclei_broad_model, tmp_path):
    _test_build_spec(unet2d_nuclei_broad_model, tmp_path / "model.zip", "torchscript", add_deepimagej_config=True)


def test_build_spec_training_data1(unet2d_nuclei_broad_model, tmp_path):
    training_data = {"id": "ilastik/stradist_dsb_training_data"}
    _test_build_spec(unet2d_nuclei_broad_model, tmp_path / "model.zip", "torchscript", training_data=training_data)


def test_build_spec_training_data2(unet2d_nuclei_broad_model, tmp_path):
    training_data = {
        "type": "dataset",
        "name": "nucleus-training-data",
        "description": "stardist nucleus training data",
        "source": "https://github.com/stardist/stardist/releases/download/0.1.0/dsb2018.zip",
    }
    _test_build_spec(unet2d_nuclei_broad_model, tmp_path / "model.zip", "torchscript", training_data=training_data)


def test_build_spec_parent1(unet2d_nuclei_broad_model, tmp_path):
    parent = {"uri": "https:/my-parent-model.org"}
    _test_build_spec(unet2d_nuclei_broad_model, tmp_path / "model.zip", "torchscript", parent=parent)


def test_build_spec_parent2(unet2d_nuclei_broad_model, tmp_path):
    parent = {"id": "10.5281/zenodo.5764892"}
    _test_build_spec(unet2d_nuclei_broad_model, tmp_path / "model.zip", "torchscript", parent=parent)


def test_build_spec_deepimagej_keras(unet2d_keras, tmp_path):
    _test_build_spec(
        unet2d_keras, tmp_path / "model.zip", "keras_hdf5", add_deepimagej_config=True, tensorflow_version="1.12"
    )


# test with original covers
def test_build_spec_with_original_covers(unet2d_nuclei_broad_model, tmp_path):
    _test_build_spec(unet2d_nuclei_broad_model, tmp_path / "model.zip", "torchscript", use_original_covers=True)


# test with absolute path for the architecture file
def test_build_spec_abs_arch_path(unet2d_nuclei_broad_model, tmp_path):
    _test_build_spec(
        unet2d_nuclei_broad_model, tmp_path / "model.zip", "pytorch_state_dict", use_absoloute_arch_path=True
    )



import warnings
from pathlib import Path
from typing import Union, Optional

import numpy as np
import torch
from numpy.testing import assert_array_almost_equal

import bioimageio.spec as spec
from bioimageio.core import load_resource_description
from bioimageio.core.resource_io import nodes
from .utils import load_model

try:
    import onnxruntime as rt
except ImportError:
    rt = None


def convert_weights_to_onnx(
    model_spec: Union[str, Path, spec.model.raw_nodes.Model],
    output_path: Union[str, Path],
    opset_version: Optional[int] = 12,
    use_tracing: bool = True,
    verbose: bool = True,
    test_decimal: int = 4
):
    """Convert model weights from format 'pytorch_state_dict' to 'onnx'.

    Args:
        model_spec: location of the resource for the input bioimageio model
        output_path: where to save the onnx weights
        opset_version: onnx opset version
        use_tracing: whether to use tracing or scripting to export the onnx format
        verbose: be verbose during the onnx export
        test_decimal: precision for testing whether the results agree
    """
    if isinstance(model_spec, (str, Path)):
        model_spec = load_resource_description(Path(model_spec))

    assert isinstance(model_spec, nodes.Model)
    with torch.no_grad():
        # load input and expected output data
        input_data = [np.load(inp).astype("float32") for inp in model_spec.test_inputs]
        input_tensors = [torch.from_numpy(inp) for inp in input_data]

        # instantiate and generate the expected output
        model = load_model(model_spec)
        expected_outputs = model(*input_tensors)
        if isinstance(expected_outputs, torch.Tensor):
            expected_outputs = [expected_outputs]
        expected_outputs = [out.numpy() for out in expected_outputs]

        if use_tracing:
            torch.onnx.export(
                model,
                input_tensors if len(input_tensors) > 1 else input_tensors[0],
                output_path,
                verbose=verbose,
                opset_version=opset_version,
                example_outputs=expected_outputs,
            )
        else:
            raise NotImplementedError

    if rt is None:
        msg = "The onnx weights were exported, but onnx rt is not available and weights cannot be checked."
        warnings.warn(msg)
        return 1

    # check the onnx model
    sess = rt.InferenceSession(str(output_path))  # does not support Path, so need to cast to str
    onnx_inputs = {input_name.name: inp for input_name, inp in zip(sess.get_inputs(), input_data)}
    outputs = sess.run(None, onnx_inputs)

    try:
        for exp, out in zip(expected_outputs, outputs):
            assert_array_almost_equal(exp, out, decimal=test_decimal)
        return 0
    except AssertionError as e:
        msg = f"The onnx weights were exported, but results before and after conversion do not agree:\n {str(e)}"
        warnings.warn(msg)
        return 1



import enum
import json
import os
import sys
from glob import glob

from pathlib import Path
from pprint import pprint
from typing import List, Optional

import typer

from bioimageio.core import __version__, prediction, commands, resource_tests, load_raw_resource_description
from bioimageio.core.prediction_pipeline import get_weight_formats
from bioimageio.spec.__main__ import app, help_version as help_version_spec
from bioimageio.spec.model.raw_nodes import WeightsFormat

try:
    from typing import get_args
except ImportError:
    from typing_extensions import get_args  # type: ignore

try:
    from bioimageio.core.weight_converter import torch as torch_converter
except ImportError:
    torch_converter = None

try:
    from bioimageio.core.weight_converter import keras as keras_converter
except ImportError:
    keras_converter = None


# extend help/version string by core version
help_version_core = f"bioimageio.core {__version__}"
help_version = f"{help_version_spec}\n{help_version_core}"
# prevent rewrapping with \b\n: https://click.palletsprojects.com/en/7.x/documentation/#preventing-rewrapping
app.info.help = "\b\n" + help_version


@app.callback()
def callback():
    typer.echo(help_version)


@app.command()
def package(
    rdf_source: str = typer.Argument(..., help="RDF source as relative file path or URI"),
    path: Path = typer.Argument(Path() / "{src_name}-package.zip", help="Save package as"),
    weights_priority_order: Optional[List[str]] = typer.Option(
        None,
        "--weights-priority-order",
        "-wpo",
        help="For model packages only. "
        "If given only the first weights matching the given weight formats are included. "
        "Defaults to include all weights present in source.",
        show_default=False,
    ),
    verbose: bool = typer.Option(False, help="show traceback of exceptions"),
):
    # typer bug: typer returns empty tuple instead of None if weights_order_priority is not given
    weights_priority_order = weights_priority_order or None

    ret_code = commands.package(
        rdf_source=rdf_source, path=path, weights_priority_order=weights_priority_order, verbose=verbose
    )
    sys.exit(ret_code)


package.__doc__ = commands.package.__doc__


# if we want to use something like "choice" for the weight formats, we need to use an enum, see:
# https://github.com/tiangolo/typer/issues/182
WeightFormatEnum = enum.Enum("WeightFormatEnum", {wf: wf for wf in get_args(WeightsFormat)})
# Enum with in values does not work with click.Choice: https://github.com/pallets/click/issues/784
# so a simple Enum with auto int values is not an option:
# WeightFormatEnum = enum.Enum("WeightFormatEnum", get_args(WeightsFormat))


@app.command()
def test_model(
    model_rdf: str = typer.Argument(
        ..., help="Path or URL to the model resource description file (rdf.yaml) or zipped model."
    ),
    weight_format: Optional[WeightFormatEnum] = typer.Option(None, help="The weight format to use."),
    devices: Optional[List[str]] = typer.Option(None, help="Devices for running the model."),
    decimal: int = typer.Option(4, help="The test precision."),
):
    # this is a weird typer bug: default devices are empty tuple although they should be None
    if len(devices) == 0:
        devices = None
    summary = resource_tests.test_model(
        model_rdf,
        weight_format=None if weight_format is None else weight_format.value,
        devices=devices,
        decimal=decimal,
    )

    if weight_format is None:
        weight_formats = get_weight_formats()
        model_weight_formats = list(load_raw_resource_description(model_rdf).weights.keys())
        for wf in weight_formats:
            if wf in model_weight_formats:
                weight_format = wf
                break
        weight_format = "unknown" if weight_format is None else weight_format

    if summary["error"] is None:
        print(f"Model test for {model_rdf} using {weight_format} weight format has passed.")
        ret_code = 0
    else:
        print(f"Model test for {model_rdf} using {weight_format} weight format has FAILED!")
        pprint(summary)
        ret_code = 1
    sys.exit(ret_code)


test_model.__doc__ = resource_tests.test_model.__doc__


@app.command()
def test_resource(
    rdf: str = typer.Argument(
        ..., help="Path or URL to the resource description file (rdf.yaml) or zipped resource package."
    ),
    weight_format: Optional[WeightFormatEnum] = typer.Option(None, help="(for model only) The weight format to use."),
    devices: Optional[List[str]] = typer.Option(None, help="(for model only) Devices for running the model."),
    decimal: int = typer.Option(4, help="(for model only) The test precision."),
):
    # this is a weird typer bug: default devices are empty tuple although they should be None
    if len(devices) == 0:
        devices = None
    summary = resource_tests.test_resource(
        rdf, weight_format=None if weight_format is None else weight_format.value, devices=devices, decimal=decimal
    )
    if summary["error"] is None:
        print(f"Resource test for {rdf} has passed.")
        ret_code = 0
    else:
        print(f"Resource test for {rdf} has FAILED!")
        pprint(summary)
        ret_code = 1
    sys.exit(ret_code)


test_resource.__doc__ = resource_tests.test_resource.__doc__


@app.command()
def predict_image(
    model_rdf: Path = typer.Argument(
        ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
    ),
    inputs: List[Path] = typer.Option(..., help="Path(s) to the model input(s)."),
    outputs: List[Path] = typer.Option(..., help="Path(s) for saveing the model output(s)."),
    # NOTE: typer currently doesn't support union types, so we only support boolean here
    # padding: Optional[Union[str, bool]] = typer.Argument(
    #     None, help="Padding to apply in each dimension passed as json encoded string."
    # ),
    # tiling: Optional[Union[str, bool]] = typer.Argument(
    #     None, help="Padding to apply in each dimension passed as json encoded string."
    # ),
    padding: Optional[bool] = typer.Option(None, help="Whether to pad the image to a size suited for the model."),
    tiling: Optional[bool] = typer.Option(None, help="Whether to run prediction in tiling mode."),
    weight_format: Optional[WeightFormatEnum] = typer.Option(None, help="The weight format to use."),
    devices: Optional[List[str]] = typer.Option(None, help="Devices for running the model."),
):

    if isinstance(padding, str):
        padding = json.loads(padding.replace("'", '"'))
        assert isinstance(padding, dict)
    if isinstance(tiling, str):
        tiling = json.loads(tiling.replace("'", '"'))
        assert isinstance(tiling, dict)

    # this is a weird typer bug: default devices are empty tuple although they should be None
    if len(devices) == 0:
        devices = None
    prediction.predict_image(
        model_rdf, inputs, outputs, padding, tiling, None if weight_format is None else weight_format.value, devices
    )


predict_image.__doc__ = prediction.predict_image.__doc__


@app.command()
def predict_images(
    model_rdf: Path = typer.Argument(
        ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
    ),
    input_pattern: str = typer.Argument(..., help="Glob pattern for the input images."),
    output_folder: str = typer.Argument(..., help="Folder to save the outputs."),
    output_extension: Optional[str] = typer.Argument(None, help="Optional output extension."),
    # NOTE: typer currently doesn't support union types, so we only support boolean here
    # padding: Optional[Union[str, bool]] = typer.Argument(
    #     None, help="Padding to apply in each dimension passed as json encoded string."
    # ),
    # tiling: Optional[Union[str, bool]] = typer.Argument(
    #     None, help="Padding to apply in each dimension passed as json encoded string."
    # ),
    padding: Optional[bool] = typer.Option(None, help="Whether to pad the image to a size suited for the model."),
    tiling: Optional[bool] = typer.Option(None, help="Whether to run prediction in tiling mode."),
    weight_format: Optional[WeightFormatEnum] = typer.Option(None, help="The weight format to use."),
    devices: Optional[List[str]] = typer.Option(None, help="Devices for running the model."),
):
    input_files = glob(input_pattern)
    input_names = [os.path.split(infile)[1] for infile in input_files]
    output_files = [os.path.join(output_folder, fname) for fname in input_names]
    if output_extension is not None:
        output_files = [f"{os.path.splitext(outfile)[0]}{output_extension}" for outfile in output_files]

    if isinstance(padding, str):
        padding = json.loads(padding.replace("'", '"'))
        assert isinstance(padding, dict)
    if isinstance(tiling, str):
        tiling = json.loads(tiling.replace("'", '"'))
        assert isinstance(tiling, dict)

    # this is a weird typer bug: default devices are empty tuple although they should be None
    if len(devices) == 0:
        devices = None
    prediction.predict_images(
        model_rdf,
        input_files,
        output_files,
        padding=padding,
        tiling=tiling,
        weight_format=None if weight_format is None else weight_format.value,
        devices=devices,
        verbose=True,
    )


predict_images.__doc__ = prediction.predict_images.__doc__


if torch_converter is not None:

    @app.command()
    def convert_torch_weights_to_onnx(
        model_rdf: Path = typer.Argument(
            ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
        ),
        output_path: Path = typer.Argument(..., help="Where to save the onnx weights."),
        opset_version: Optional[int] = typer.Argument(12, help="Onnx opset version."),
        use_tracing: bool = typer.Option(True, help="Whether to use torch.jit tracing or scripting."),
        verbose: bool = typer.Option(True, help="Verbosity"),
    ):
        ret_code = torch_converter.convert_weights_to_onnx(model_rdf, output_path, opset_version, use_tracing, verbose)
        sys.exit(ret_code)

    convert_torch_weights_to_onnx.__doc__ = torch_converter.convert_weights_to_onnx.__doc__

    @app.command()
    def convert_torch_weights_to_torchscript(
        model_rdf: Path = typer.Argument(
            ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
        ),
        output_path: Path = typer.Argument(..., help="Where to save the torchscript weights."),
        use_tracing: bool = typer.Option(True, help="Whether to use torch.jit tracing or scripting."),
    ):
        ret_code = torch_converter.convert_weights_to_torchscript(model_rdf, output_path, use_tracing)
        sys.exit(ret_code)

    convert_torch_weights_to_torchscript.__doc__ = torch_converter.convert_weights_to_torchscript.__doc__


if keras_converter is not None:

    @app.command()
    def convert_keras_weights_to_tensorflow(
        model_rdf: Path = typer.Argument(
            ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
        ),
        output_path: Path = typer.Argument(..., help="Where to save the tensorflow weights."),
    ):
        ret_code = keras_converter.convert_weights_to_tensorflow_saved_model_bundle(model_rdf, output_path)
        sys.exit(ret_code)

    convert_keras_weights_to_tensorflow.__doc__ = (
        keras_converter.convert_weights_to_tensorflow_saved_model_bundle.__doc__
    )


if __name__ == "__main__":
    app()



from dataclasses import dataclass, field, fields
from typing import Any, Dict, Optional, Sequence, Set, Union

import numpy as np
import xarray as xr

from bioimageio.core.statistical_measures import Mean, Measure, Percentile, Std

try:
    from typing import Literal, get_args
except ImportError:
    from typing_extensions import Literal, get_args  # type: ignore


def _get_fixed(
    fixed: Union[float, Sequence[float]], tensor: xr.DataArray, axes: Optional[Sequence[str]]
) -> Union[float, xr.DataArray]:
    if axes is None:
        return fixed

    fixed_shape = tuple(s for d, s in tensor.sizes.items() if d not in axes)
    fixed_dims = tuple(d for d in tensor.dims if d not in axes)
    fixed = np.array(fixed).reshape(fixed_shape)
    return xr.DataArray(fixed, dims=fixed_dims)


@dataclass
class Processing:
    """base class for all Pre- and Postprocessing transformations"""

    tensor_name: str
    computed_dataset_statistics: Dict[str, Dict[Measure, Any]] = field(init=False)
    computed_sample_statistics: Dict[str, Dict[Measure, Any]] = field(init=False)

    def get_required_dataset_statistics(self) -> Dict[str, Set[Measure]]:
        """
        Specifies which dataset measures are required from what tensor.
        Returns: dataset measures required to apply this processing indexed by <tensor_name>.
        """
        return {}

    def get_required_sample_statistics(self) -> Dict[str, Set[Measure]]:
        """
        Specifies which sample measures are required from what tensor.
        Returns: sample measures required to apply this processing indexed by <tensor_name>.
        """
        return {}

    def set_computed_dataset_statistics(self, computed: Dict[str, Dict[Measure, Any]]):
        """helper to set computed statistics and check if they match the requirements"""
        for tensor_name, req_measures in self.get_required_dataset_statistics().items():
            comp_measures = computed.get(tensor_name, {})
            for req_measure in req_measures:
                if req_measure not in comp_measures:
                    raise ValueError(f"Missing required measure {req_measure} for {tensor_name}")
        self.computed_dataset_statistics = computed

    def set_computed_sample_statistics(self, computed: Dict[str, Dict[Measure, Any]]):
        """helper to set computed statistics and check if they match the requirements"""
        for tensor_name, req_measures in self.get_required_sample_statistics().items():
            comp_measures = computed.get(tensor_name, {})
            for req_measure in req_measures:
                if req_measure not in comp_measures:
                    raise ValueError(f"Missing required measure {req_measure} for {tensor_name}")
        self.computed_sample_statistics = computed

    def get_computed_dataset_statistics(self, tensor_name: str, measure: Measure):
        """helper to unpack self.computed_dataset_statistics"""
        ret = self.computed_dataset_statistics.get(tensor_name, {}).get(measure)
        if ret is None:
            raise RuntimeError(f"Missing computed {measure} for {tensor_name} dataset.")

        return ret

    def get_computed_sample_statistics(self, tensor_name: str, measure: Measure):
        """helper to unpack self.computed_sample_statistics"""
        ret = self.computed_sample_statistics.get(tensor_name, {}).get(measure)
        if ret is None:
            raise RuntimeError(f"Missing computed {measure} for {tensor_name} sample.")

        return ret

    def __call__(self, tensor: xr.DataArray) -> xr.DataArray:
        return self.apply(tensor)

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        """apply processing to named tensors"""
        raise NotImplementedError

    def __post_init__(self):
        """validate common kwargs by their annotations"""
        self.computed_dataset_statistics = {}
        self.computed_sample_statistics = {}

        for f in fields(self):
            if f.name == "mode":
                assert hasattr(self, "mode")
                if self.mode not in get_args(f.type):
                    raise NotImplementedError(
                        f"Unsupported mode {self.mode} for {self.__class__.__name__}: {self.mode}"
                    )


#
# helpers
#
def ensure_dtype(tensor: xr.DataArray, *, dtype) -> xr.DataArray:
    """
    Convert array to a given datatype
    """
    return tensor.astype(dtype)


#
# Pre- and Postprocessing implementations
#


@dataclass
class Binarize(Processing):
    threshold: float

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return ensure_dtype(tensor > self.threshold, dtype="float32")


@dataclass
class Clip(Processing):
    min: float
    max: float

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return ensure_dtype(tensor.clip(min=self.min, max=self.max), dtype="float32")


@dataclass
class EnsureDtype(Processing):
    dtype: str

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return ensure_dtype(tensor, dtype=self.dtype)


@dataclass
class ScaleLinear(Processing):
    """scale the tensor with a fixed multiplicative and additive factor"""

    gain: Union[float, Sequence[float]]
    offset: Union[float, Sequence[float]]
    axes: Optional[Sequence[str]] = None

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        scale_axes = tuple(ax for ax in tensor.dims if (ax not in self.axes and ax != "b"))
        if scale_axes:
            gain = xr.DataArray(np.atleast_1d(self.gain), dims=scale_axes)
            offset = xr.DataArray(np.atleast_1d(self.offset), dims=scale_axes)
        else:
            gain = self.gain
            offset = self.offset

        return ensure_dtype(tensor * gain + offset, dtype="float32")

    def __post_init__(self):
        super().__post_init__()
        if self.axes is None:
            assert isinstance(self.gain, (int, float))
            assert isinstance(self.offset, (int, float))


@dataclass
class ScaleMeanVariance(Processing):
    ...


@dataclass
class ScaleRange(Processing):
    mode: Literal["per_sample", "per_dataset"] = "per_sample"
    axes: Optional[Sequence[str]] = None
    min_percentile: float = 0.0
    max_percentile: float = 100.0
    eps: float = 1e-6
    reference_tensor: Optional[str] = None

    def get_required_dataset_statistics(self) -> Dict[str, Set[Measure]]:
        if self.mode == "per_sample":
            return {}
        elif self.mode == "per_dataset":
            measures = {
                Percentile(self.min_percentile, axes=self.axes),
                Percentile(self.max_percentile, axes=self.axes),
            }
            return {self.reference_tensor or self.tensor_name: measures}
        else:
            raise ValueError(self.mode)

    def get_required_sample_statistics(self) -> Dict[str, Set[Measure]]:
        if self.mode == "per_sample":
            measures = {
                Percentile(self.min_percentile, axes=self.axes),
                Percentile(self.max_percentile, axes=self.axes),
            }
            return {self.reference_tensor or self.tensor_name: measures}
        elif self.mode == "per_dataset":
            return {}
        else:
            raise ValueError(self.mode)

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        ref_name = self.reference_tensor or self.tensor_name
        if self.axes:
            axes = tuple(self.axes)
        else:
            axes = None

        if self.mode == "per_sample":
            get_stat = self.get_computed_sample_statistics
        elif self.mode == "per_dataset":
            get_stat = self.get_computed_dataset_statistics
        else:
            raise ValueError(self.mode)

        v_lower = get_stat(ref_name, Percentile(self.min_percentile, axes=axes))
        v_upper = get_stat(ref_name, Percentile(self.max_percentile, axes=axes))

        return ensure_dtype((tensor - v_lower) / (v_upper - v_lower + self.eps), dtype="float32")

    def __post_init__(self):
        super().__post_init__()
        self.axes = None if self.axes is None else tuple(self.axes)  # make sure axes is Tuple[str] or None


@dataclass
class Sigmoid(Processing):
    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return 1.0 / (1.0 + xr.ufuncs.exp(-tensor))


@dataclass
class ZeroMeanUnitVariance(Processing):
    mode: Literal["fixed", "per_sample", "per_dataset"] = "per_sample"
    mean: Optional[Union[float, Sequence[float]]] = None
    std: Optional[Union[float, Sequence[float]]] = None
    axes: Optional[Sequence[str]] = None
    eps: float = 1.0e-6

    def get_required_dataset_statistics(self) -> Dict[str, Set[Measure]]:
        if self.mode == "per_dataset":
            return {self.tensor_name: {Mean(), Std()}}
        else:
            return {}

    def get_required_sample_statistics(self) -> Dict[str, Set[Measure]]:
        if self.mode == "per_sample":
            return {self.tensor_name: {Mean(), Std()}}
        else:
            return {}

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        axes = None if self.axes is None else tuple(self.axes)
        if self.mode == "fixed":
            assert self.mean is not None and self.std is not None
            mean = _get_fixed(self.mean, tensor, axes)
            std = _get_fixed(self.std, tensor, axes)
        elif self.mode == "per_sample":
            mean = Mean(axes).compute(tensor)
            std = Std(axes).compute(tensor)
        elif self.mode == "per_dataset":
            mean = self.get_computed_dataset_statistics(self.tensor_name, Mean(axes))
            std = self.get_computed_dataset_statistics(self.tensor_name, Std(axes))
        else:
            raise ValueError(self.mode)

        tensor = (tensor - mean) / (std + self.eps)
        return ensure_dtype(tensor, dtype="float32")



import bioimageio.spec
from .io_ import (
    export_resource_package,
    load_resource_description,
    save_raw_resource_description,
    serialize_raw_resource_description,
)

load_raw_resource_description = bioimageio.spec.load_raw_resource_description



from dataclasses import dataclass
from typing import Optional, Tuple

import xarray as xr


@dataclass(frozen=True)
class Measure:
    def compute(self, tensor: xr.DataArray):
        raise NotImplementedError(self.__class__.__name__)


@dataclass(frozen=True)
class Mean(Measure):
    axes: Optional[Tuple[str]] = None

    def compute(self, tensor: xr.DataArray) -> xr.DataArray:
        return tensor.mean(dim=self.axes)


@dataclass(frozen=True)
class Std(Measure):
    axes: Optional[Tuple[str]] = None

    def compute(self, tensor: xr.DataArray) -> xr.DataArray:
        return tensor.std(dim=self.axes)


@dataclass(frozen=True)
class Percentile(Measure):
    n: float
    axes: Optional[Tuple[str]] = None

    def __post_init__(self):
        assert self.n >= 0
        assert self.n <= 100

    def compute(self, tensor: xr.DataArray) -> xr.DataArray:
        return tensor.quantile(self.n / 100.0, dim=self.axes)
